@@ -34,7 +34,7 @@
 #define ARRAY_SIZE(a) (sizeof (a) / sizeof (a)[0])
 
 
-char *id_str(snd_ctl_elem_id_t *id)
+static char *id_str(snd_ctl_elem_id_t *id)
 {
 	static char str[128];
 	assert(id);
@@ -47,7 +47,7 @@ char *id_str(snd_ctl_elem_id_t *id)
 	return str;
 }
 
-char *num_str(long n)
+static char *num_str(long n)
 {
 	static char str[32];
 	sprintf(str, "%ld", n);
@@ -58,7 +58,7 @@ static int snd_config_integer_add(snd_config_t *father, char *id, long integer)
 {
 	int err;
 	snd_config_t *leaf;
-	err = snd_config_make_integer(&leaf, id);
+	err = snd_config_imake_integer(&leaf, id, integer);
 	if (err < 0)
 		return err;
 	err = snd_config_add(father, leaf);
@@ -66,11 +66,6 @@ static int snd_config_integer_add(snd_config_t *father, char *id, long integer)
 		snd_config_delete(leaf);
 		return err;
 	}
-	err = snd_config_set_integer(leaf, integer);
-	if (err < 0) {
-		snd_config_delete(leaf);
-		return err;
-	}
 	return 0;
 }
 
@@ -78,7 +73,7 @@ static int snd_config_integer64_add(snd_config_t *father, char *id, long long in
 {
 	int err;
 	snd_config_t *leaf;
-	err = snd_config_make_integer64(&leaf, id);
+	err = snd_config_imake_integer64(&leaf, id, integer);
 	if (err < 0)
 		return err;
 	err = snd_config_add(father, leaf);
@@ -86,11 +81,6 @@ static int snd_config_integer64_add(snd_config_t *father, char *id, long long in
 		snd_config_delete(leaf);
 		return err;
 	}
-	err = snd_config_set_integer64(leaf, integer);
-	if (err < 0) {
-		snd_config_delete(leaf);
-		return err;
-	}
 	return 0;
 }
 
@@ -98,7 +88,7 @@ static int snd_config_string_add(snd_config_t *father, const char *id, const cha
 {
 	int err;
 	snd_config_t *leaf;
-	err = snd_config_make_string(&leaf, id);
+	err = snd_config_imake_string(&leaf, id, string);
 	if (err < 0)
 		return err;
 	err = snd_config_add(father, leaf);
@@ -106,11 +96,6 @@ static int snd_config_string_add(snd_config_t *father, const char *id, const cha
 		snd_config_delete(leaf);
 		return err;
 	}
-	err = snd_config_set_string(leaf, string);
-	if (err < 0) {
-		snd_config_delete(leaf);
-		return err;
-	}
 	return 0;
 }
 
@@ -179,14 +164,18 @@ static unsigned int *str_to_tlv(const char *s)
 }
 
 /*
- * add the TLV string and dB ranges to comment fields
+ * add the TLV string, dB ranges, and dB values to comment fields
  */
 static int add_tlv_comments(snd_ctl_t *handle, snd_ctl_elem_id_t *id,
-			    snd_ctl_elem_info_t *info, snd_config_t *comment)
+			    snd_ctl_elem_info_t *info, snd_ctl_elem_value_t *ctl,
+			    snd_config_t *comment)
 {
 	unsigned int tlv[MAX_USER_TLV_SIZE];
 	unsigned int *db;
-	long dbmin, dbmax;
+	long rangemin, rangemax;
+	long dbmin, dbmax, dbgain;
+	snd_config_t *value;
+	unsigned int i, count;
 	int err;
 
 	if (snd_ctl_elem_tlv_read(handle, id, tlv, sizeof(tlv)) < 0)
@@ -208,13 +197,35 @@ static int add_tlv_comments(snd_ctl_t *handle, snd_ctl_elem_id_t *id,
 	if (err <= 0)
 		return 0;
 
-	snd_tlv_get_dB_range(db, snd_ctl_elem_info_get_min(info),
-			     snd_ctl_elem_info_get_max(info),
-			     &dbmin, &dbmax);
+	rangemin = snd_ctl_elem_info_get_min(info);
+	rangemax = snd_ctl_elem_info_get_max(info);
+	snd_tlv_get_dB_range(db, rangemin, rangemax, &dbmin, &dbmax);
 	if (err < 0)
 		return err;
 	snd_config_integer_add(comment, "dbmin", dbmin);
 	snd_config_integer_add(comment, "dbmax", dbmax);
+
+	if (snd_ctl_elem_info_get_type(info) == SND_CTL_ELEM_TYPE_INTEGER) {
+		err = snd_config_compound_add(comment, "dbvalue", 1, &value);
+		if (err < 0) {
+			error("snd_config_compound_add: %s", snd_strerror(err));
+			return err;
+		}
+		count = snd_ctl_elem_info_get_count(info);
+		for (i = 0; i < count; i++) {
+			err = snd_tlv_convert_to_dB(db, rangemin, rangemax,
+					snd_ctl_elem_value_get_integer(ctl, i), &dbgain);
+			if (err < 0) {
+				error("snd_tlv_convert_to_dB: %s", snd_strerror(err));
+				return err;
+			}
+			err = snd_config_integer_add(value, num_str(i), dbgain);
+			if (err < 0) {
+				error("snd_config_integer_add: %s", snd_strerror(err));
+				return err;
+			}
+		}
+	}
 	return 0;
 }
 
@@ -240,8 +251,7 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 		return err;
 	}
 
-	if (snd_ctl_elem_info_is_inactive(info) ||
-				!snd_ctl_elem_info_is_readable(info))
+	if (!snd_ctl_elem_info_is_readable(info))
 		return 0;
 	snd_ctl_elem_value_set_id(ctl, id);
 	err = snd_ctl_elem_read(handle, ctl);
@@ -255,9 +265,9 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 		error("snd_config_compound_add: %s", snd_strerror(err));
 		return err;
 	}
-	err = snd_config_compound_add(control, "comment", 1, &comment);
+	err = snd_config_make_compound(&comment, "comment", 0);
 	if (err < 0) {
-		error("snd_config_compound_add: %s", snd_strerror(err));
+		error("snd_config_make_compound: %s", snd_strerror(err));
 		return err;
 	}
 
@@ -317,7 +327,7 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 			return err;
 		}
 		if (snd_ctl_elem_info_is_tlv_readable(info)) {
-			err = add_tlv_comments(handle, id, info, comment);
+			err = add_tlv_comments(handle, id, info, ctl, comment);
 			if (err < 0)
 				return err;
 		}
@@ -421,7 +431,7 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 			error("snd_config_string_add: %s", snd_strerror(err));
 			return err;
 		}
-		return 0;
+		goto finish;
 	}
 	default:
 		break;
@@ -435,21 +445,21 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 				error("snd_config_string_add: %s", snd_strerror(err));
 				return err;
 			}
-			return 0;
+			goto finish;
 		case SND_CTL_ELEM_TYPE_INTEGER:
 			err = snd_config_integer_add(control, "value", snd_ctl_elem_value_get_integer(ctl, 0));
 			if (err < 0) {
 				error("snd_config_integer_add: %s", snd_strerror(err));
 				return err;
 			}
-			return 0;
+			goto finish;
 		case SND_CTL_ELEM_TYPE_INTEGER64:
 			err = snd_config_integer64_add(control, "value", snd_ctl_elem_value_get_integer64(ctl, 0));
 			if (err < 0) {
 				error("snd_config_integer64_add: %s", snd_strerror(err));
 				return err;
 			}
-			return 0;
+			goto finish;
 		case SND_CTL_ELEM_TYPE_ENUMERATED:
 		{
 			unsigned int v = snd_ctl_elem_value_get_enumerated(ctl, 0);
@@ -464,7 +474,7 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 			}
 			if (err < 0)
 				error("snd_config add: %s", snd_strerror(err));
-			return 0;
+			goto finish;
 		}
 		default:
 			error("Unknown control type: %d\n", type);
@@ -529,6 +539,12 @@ static int get_control(snd_ctl_t *handle, snd_ctl_elem_id_t *id, snd_config_t *t
 		return -EINVAL;
 	}
 	
+finish:
+	err = snd_config_add(control, comment);
+	if (err < 0) {
+		error("snd_config_add: %s", snd_strerror(err));
+		return err;
+	}
 	return 0;
 }
 	
@@ -538,6 +554,7 @@ static int get_controls(int cardno, snd_config_t *top)
 	snd_ctl_card_info_t *info;
 	snd_config_t *state, *card, *control;
 	snd_ctl_elem_list_t *list;
+	snd_ctl_elem_id_t *elem_id;
 	unsigned int idx;
 	int err;
 	char name[32];
@@ -545,6 +562,7 @@ static int get_controls(int cardno, snd_config_t *top)
 	const char *id;
 	snd_ctl_card_info_alloca(&info);
 	snd_ctl_elem_list_alloca(&list);
+	snd_ctl_elem_id_alloca(&elem_id);
 
 	sprintf(name, "hw:%d", cardno);
 	err = snd_ctl_open(&handle, name, SND_CTL_READONLY);
@@ -600,10 +618,6 @@ static int get_controls(int cardno, snd_config_t *top)
 		goto _close;
 	}
 	count = snd_ctl_elem_list_get_count(list);
-	if (count < 0) {
-		err = 0;
-		goto _close;
-	}
 	err = snd_config_compound_add(card, "control", count > 0, &control);
 	if (err < 0) {
 		error("snd_config_compound_add: %s", snd_strerror(err));
@@ -623,10 +637,8 @@ static int get_controls(int cardno, snd_config_t *top)
 		goto _free;
 	}
 	for (idx = 0; idx < count; ++idx) {
-		snd_ctl_elem_id_t *id;
-		snd_ctl_elem_id_alloca(&id);
-		snd_ctl_elem_list_get_id(list, idx, id);
-		err = get_control(handle, id, control);
+		snd_ctl_elem_list_get_id(list, idx, elem_id);
+		err = get_control(handle, elem_id, control);
 		if (err < 0)
 			goto _free;
 	}		
@@ -765,7 +777,7 @@ static int config_integer64(snd_config_t *n, long long *val, int doit)
 	return err;
 }
 
-static int is_user_control(snd_config_t *conf)
+static int check_comment_access(snd_config_t *conf, const char *str)
 {
 	snd_config_iterator_t i, next;
 
@@ -777,7 +789,7 @@ static int is_user_control(snd_config_t *conf)
 		if (strcmp(id, "access") == 0) {
 			if (snd_config_get_string(n, &s) < 0)
 				return 0;
-			if (strstr(s, "user"))
+			if (strstr(s, str))
 				return 1;
 		}
 	}
@@ -918,32 +930,15 @@ static int add_user_control(snd_ctl_t *handle, snd_ctl_elem_info_t *info, snd_co
 	return snd_ctl_elem_info(handle, info);
 }
 
-/*
- * look for a config node with the given item name
- */
-static snd_config_t *search_comment_item(snd_config_t *conf, const char *name)
-{
-	snd_config_iterator_t i, next;
-	snd_config_for_each(i, next, conf) {
-		snd_config_t *n = snd_config_iterator_entry(i);
-		const char *id;
-		if (snd_config_get_id(n, &id) < 0)
-			continue;
-		if (strcmp(id, name) == 0)
-			return n;
-	}
-	return NULL;
-}
-
 /*
  * check whether the config item has the same of compatible type
  */
 static int check_comment_type(snd_config_t *conf, int type)
 {
-	snd_config_t *n = search_comment_item(conf, "type");
+	snd_config_t *n;
 	int ctype;
 
-	if (!n)
+	if (snd_config_search(conf, "type", &n) < 0)
 		return 0; /* not defined */
 	ctype = get_comment_type(n);
 	if (ctype == type)
@@ -966,20 +961,31 @@ static int check_comment_type(snd_config_t *conf, int type)
 static int convert_to_new_db(snd_config_t *value, long omin, long omax,
 			     long nmin, long nmax,
 			     long odbmin, long odbmax,
-			     long ndbmin, long ndbmax,
+			     snd_config_t *comment, const char *index,
+			     snd_ctl_t *device, snd_ctl_elem_id_t *id,
 			     int doit)
 {
-	long val;
-	if (config_integer(value, &val, doit) < 0)
-		return -EINVAL;
-	if (val < omin || val > omax)
-		return -EINVAL;
-	val = ((val - omin) * (odbmax - odbmin)) / (omax - omin) + odbmin;
-	if (val < ndbmin)
-		val = ndbmin;
-	else if (val > ndbmax)
-		val = ndbmax;
-	val = ((val - ndbmin) * (nmax - nmin)) / (ndbmax - ndbmin) + nmin;
+	snd_config_t *db_node;
+	long db, val;
+	int err;
+
+	if (snd_config_searchv(comment, &db_node, "dbvalue", index, NULL) < 0 ||
+	    snd_config_get_integer(db_node, &db) < 0) {
+		err = config_integer(value, &val, doit);
+		if (err < 0)
+			return err;
+		if (val < omin || val > omax)
+			return -EINVAL;
+		db = ((val - omin) * (odbmax - odbmin)) / (omax - omin) + odbmin;
+	}
+
+	err = snd_ctl_convert_from_dB(device, id, db, &val, db > 0);
+	if (err < 0)
+		return err;
+	if (val < nmin)
+		val = nmin;
+	else if (val > nmax)
+		val = nmax;
 	return snd_config_set_integer(value, val);
 }
 
@@ -999,8 +1005,7 @@ static int check_comment_range(snd_ctl_t *handle, snd_config_t *conf,
 	long ndbmin, ndbmax;
 	snd_ctl_elem_id_t *id;
 
-	n = search_comment_item(conf, "range");
-	if (!n)
+	if (snd_config_search(conf, "range", &n) < 0)
 		return 0;
 	if (get_comment_range(n, SND_CTL_ELEM_TYPE_INTEGER,
 			      &omin, &omax, &ostep) < 0)
@@ -1015,13 +1020,11 @@ static int check_comment_range(snd_ctl_t *handle, snd_config_t *conf,
 	if (omin >= omax || nmin >= nmax)
 		return 0; /* invalid values */
 
-	n = search_comment_item(conf, "dbmin");
-	if (!n)
+	if (snd_config_search(conf, "dbmin", &n) < 0)
 		return 0;
 	if (config_integer(n, &odbmin, doit) < 0)
 		return 0;
-	n = search_comment_item(conf, "dbmax");
-	if (!n)
+	if (snd_config_search(conf, "dbmax", &n) < 0)
 		return 0;
 	if (config_integer(n, &odbmax, doit) < 0)
 		return 0;
@@ -1042,12 +1045,17 @@ static int check_comment_range(snd_ctl_t *handle, snd_config_t *conf,
 		snd_config_iterator_t i, next;
 		snd_config_for_each(i, next, value) {
 			snd_config_t *n = snd_config_iterator_entry(i);
+			const char *idxstr;
+			if (snd_config_get_id(n, &idxstr) < 0)
+				continue;
 			convert_to_new_db(n, omin, omax, nmin, nmax,
-					  odbmin, odbmax, ndbmin, ndbmax, doit);
+					  odbmin, odbmax, conf, idxstr,
+					  handle, id, doit);
 		}
 	} else
 		convert_to_new_db(value, omin, omax, nmin, nmax,
-				  odbmin, odbmax, ndbmin, ndbmax, doit);
+				  odbmin, odbmax, conf, "0",
+				  handle, id, doit);
 	return 0;
 }
 
@@ -1121,7 +1129,6 @@ static int restore_config_value2(snd_ctl_t *handle, snd_ctl_elem_info_t *info,
 		}
 		snd_ctl_elem_value_set_byte(ctl, idx, val);
 		return 1;
-		break;
 	default:
 		break;
 	}
@@ -1166,7 +1173,7 @@ static int set_control(snd_ctl_t *handle, snd_config_t *control,
 		return -EINVAL;
 	}
 	numid = atoi(id);
-	if (numid > *maxnumid)
+	if ((int)numid > *maxnumid)
 		*maxnumid = numid;
 	snd_config_for_each(i, next, control) {
 		snd_config_t *n = snd_config_iterator_entry(i);
@@ -1183,10 +1190,6 @@ static int set_control(snd_ctl_t *handle, snd_config_t *control,
 		}
 		if (strcmp(fld, "iface") == 0) {
 			iface = (snd_ctl_elem_iface_t)config_iface(n);
-			if (iface < 0) {
-				cerror(doit, "control.%d.%s is invalid", numid, fld);
-				return -EINVAL;
-			}
 			continue;
 		}
 		if (strcmp(fld, "device") == 0) {
@@ -1243,22 +1246,20 @@ static int set_control(snd_ctl_t *handle, snd_config_t *control,
 		snd_ctl_elem_info_set_numid(info, numid);
 		err = snd_ctl_elem_info(handle, info);
 	}
-	if (err < 0) {
-		if (iface >= 0 && name) {
-			snd_ctl_elem_info_set_numid(info, 0);
-			snd_ctl_elem_info_set_interface(info, iface);
-			snd_ctl_elem_info_set_device(info, device);
-			snd_ctl_elem_info_set_subdevice(info, subdevice);
-			snd_ctl_elem_info_set_name(info, name);
-			snd_ctl_elem_info_set_index(info, index);
-			err = snd_ctl_elem_info(handle, info);
-			if (err < 0 && comment && is_user_control(comment)) {
-				err = add_user_control(handle, info, comment);
-				if (err < 0) {
-					cerror(doit, "failed to add user control #%d (%s)",
-					      numid, snd_strerror(err));
-					return err;
-				}
+	if (err < 0 && name) {
+		snd_ctl_elem_info_set_numid(info, 0);
+		snd_ctl_elem_info_set_interface(info, iface);
+		snd_ctl_elem_info_set_device(info, device);
+		snd_ctl_elem_info_set_subdevice(info, subdevice);
+		snd_ctl_elem_info_set_name(info, name);
+		snd_ctl_elem_info_set_index(info, index);
+		err = snd_ctl_elem_info(handle, info);
+		if (err < 0 && comment && check_comment_access(comment, "user")) {
+			err = add_user_control(handle, info, comment);
+			if (err < 0) {
+				cerror(doit, "failed to add user control #%d (%s)",
+				       numid, snd_strerror(err));
+				return err;
 			}
 		}
 	}
@@ -1302,6 +1303,9 @@ static int set_control(snd_ctl_t *handle, snd_config_t *control,
 				return -EINVAL;
 			}
 		}
+		/* inactive controls are not restored */
+		if (comment && check_comment_access(comment, "inactive"))
+			return 0;
 	}
 
 	if (snd_ctl_elem_info_is_inactive(info) ||
@@ -1372,8 +1376,7 @@ static int set_control(snd_ctl_t *handle, snd_config_t *control,
 		if (snd_config_get_id(n, &id) < 0)
 			continue;
 		idx = atoi(id);
-		if (idx < 0 || idx >= count || 
-		    set[idx]) {
+		if (idx >= count || set[idx]) {
 			cerror(doit, "bad control.%d.value index", numid);
 			if (!force_restore || !doit)
 				return -EINVAL;
@@ -1409,12 +1412,13 @@ static int set_controls(int card, snd_config_t *top, int doit)
 	snd_ctl_card_info_t *info;
 	snd_config_t *control;
 	snd_config_iterator_t i, next;
-	int err, maxnumid;
+	int err, maxnumid = -1;
 	char name[32], tmpid[16];
 	const char *id;
 	snd_ctl_card_info_alloca(&info);
 
 	sprintf(name, "hw:%d", card);
+	dbg("device='%s', doit=%i", name, doit);
 	err = snd_ctl_open(&handle, name, 0);
 	if (err < 0) {
 		error("snd_ctl_open error: %s", snd_strerror(err));
@@ -1426,6 +1430,7 @@ static int set_controls(int card, snd_config_t *top, int doit)
 		goto _close;
 	}
 	id = snd_ctl_card_info_get_id(info);
+	dbg("card-info-id: '%s'", id);
 	err = snd_config_searchv(top, &control, "state", id, "control", 0);
 	if (err < 0) {
 		if (force_restore) {
@@ -1451,24 +1456,25 @@ static int set_controls(int card, snd_config_t *top, int doit)
 			goto _close;
 	}
 
+	dbg("maxnumid=%i", maxnumid);
 	/* check if we have additional controls in driver */
 	/* in this case we should go through init procedure */
-	if (!doit) {
-		snd_ctl_elem_id_t *id;
+	if (!doit && maxnumid >= 0) {
 		snd_ctl_elem_info_t *info;
-		snd_ctl_elem_id_alloca(&id);
 		snd_ctl_elem_info_alloca(&info);
 		snd_ctl_elem_info_set_numid(info, maxnumid+1);
 		if (snd_ctl_elem_info(handle, info) == 0) {
 			/* not very informative */
 			/* but value is used for check only */
 			err = -EAGAIN;
+			dbg("more controls than maxnumid?");
 			goto _close;
 		}
 	}
 
  _close:
 	snd_ctl_close(handle);
+	dbg("result code: %i", err);
 	return err;
 }
 
@@ -1548,7 +1554,8 @@ int save_state(const char *file, const char *cardname)
 	return 0;
 }
 
-int load_state(const char *file, const char *initfile, const char *cardname)
+int load_state(const char *file, const char *initfile, const char *cardname,
+	       int do_init)
 {
 	int err, finalerr = 0;
 	snd_config_t *config;
@@ -1586,15 +1593,17 @@ int load_state(const char *file, const char *initfile, const char *cardname)
 			if (card < 0)
 				break;
 			first = 0;
+			if (!do_init)
+				break;
 			sprintf(cardname1, "%i", card);
 			err = init(initfile, cardname1);
 			if (err < 0) {
 				finalerr = err;
-				initfailed(card, "init");
+				initfailed(card, "init", err);
 			}
-			initfailed(card, "restore");
+			initfailed(card, "restore", -ENOENT);
 		}
-		if (!first)
+		if (first)
 			finalerr = 0;	/* no cards, no error code */
 		return finalerr;
 	}
@@ -1621,18 +1630,18 @@ int load_state(const char *file, const char *initfile, const char *cardname)
 			}
 			first = 0;
 			/* do a check if controls matches state file */
- 			if (set_controls(card, config, 0)) {
+ 			if (do_init && set_controls(card, config, 0)) {
 				sprintf(cardname1, "%i", card);
 				err = init(initfile, cardname1);
 				if (err < 0) {
-					initfailed(card, "init");
+					initfailed(card, "init", err);
 					finalerr = err;
 				}
 			}
 			if ((err = set_controls(card, config, 1))) {
 				if (!force_restore)
 					finalerr = err;
-				initfailed(card, "restore");
+				initfailed(card, "restore", err);
 			}
 		}
 	} else {
@@ -1644,15 +1653,15 @@ int load_state(const char *file, const char *initfile, const char *cardname)
 			return -ENODEV;
 		}
 		/* do a check if controls matches state file */
-		if (set_controls(cardno, config, 0)) {
+		if (do_init && set_controls(cardno, config, 0)) {
 			err = init(initfile, cardname);
 			if (err < 0) {
-				initfailed(cardno, "init");
-				return err;
+				initfailed(cardno, "init", err);
+				finalerr = err;
 			}
 		}
 		if ((err = set_controls(cardno, config, 1))) {
-			initfailed(cardno, "restore");
+			initfailed(cardno, "restore", err);
 			if (!force_restore)
 				return err;
 		}
